-------------------------------------------------------------------------------------------------------------------------------------------------------------
-- VEAF spawnable aircrafts editor tool for DCS World
-- By Zip (2021)
--
-- Features:
-- ---------
-- * This tool processes a mission and update flight plans.
-- * The flight plans templates can be customized
--
-- Prerequisite:
-- ------------
-- * The mission file archive must already be exploded ; the script only works on the mission files, not directly on the .miz archive
--
-- Basic Usage:
-- ------------
-- Call the script by running it in a lua environment ; it needs the veafMissionEditor library, so the script working directory must contain the veafMissionEditor.lua file
-- 
-- veafSpawnableAircraftsEditor.lua <mission folder path> <settings file> [-debug|-trace]
-- 
-- Command line options:
-- * <mission folder path> the path to the exploded mission files (no trailing backslash)
-- * <settings file> the path to the settings file
-- * -debug if set, the script will output some information ; useful to find out which units were edited
-- * -trace if set, the script will output a lot of information : useful to understand what went wrong
-- * -import if set, the script will import data from the .miz file (the exploded mission folder) instead of injecting data. Useful to update the settings file
-- * -nameFilter, a regex that will be used to filter the groups that are processed (in either direction) by matching their names. Default to nil, all groups processed
-------------------------------------------------------------------------------------------------------------------------------------------------------------

require("src/scripts/veaf/veafMissionEditor")

veafSpawnableAircraftsEditor = {}

-------------------------------------------------------------------------------------------------------------------------------------------------------------
-- Global settings. Stores the script constants
-------------------------------------------------------------------------------------------------------------------------------------------------------------

--- Identifier. All output in the log will start with this.
veafSpawnableAircraftsEditor.Id = "SPAWN_AC_EDITOR - "

--- Version.
veafSpawnableAircraftsEditor.Version = "1.0.0"

-- trace level, specific to this module
veafSpawnableAircraftsEditor.Trace = false
veafSpawnableAircraftsEditor.Debug = false

-- default position for all flights that are processed (imported from .miz to the configuration file) in meters, DCS model
local DEFAULT_POSITION_X = 0
local DEFAULT_POSITION_Y = 0

-------------------------------------------------------------------------------------------------------------------------------------------------------------
-- Do not change anything below unless you know what you are doing!
-------------------------------------------------------------------------------------------------------------------------------------------------------------

local CATEGORIES = {"plane", "helicopter"}

-------------------------------------------------------------------------------------------------------------------------------------------------------------
-- Utility methods
-------------------------------------------------------------------------------------------------------------------------------------------------------------

function veafSpawnableAircraftsEditor.logError(message)
    print(veafSpawnableAircraftsEditor.Id .. message)
end

function veafSpawnableAircraftsEditor.logInfo(message)
    print(veafSpawnableAircraftsEditor.Id .. message)
end

function veafSpawnableAircraftsEditor.logDebug(message)
  if message and veafSpawnableAircraftsEditor.Debug then
    print(veafSpawnableAircraftsEditor.Id .. message)
  end
end

function veafSpawnableAircraftsEditor.logTrace(message)
  if message and veafSpawnableAircraftsEditor.Trace then
    print(veafSpawnableAircraftsEditor.Id .. message)
  end
end

local function _p(o, level)
  local MAX_LEVEL = 20
  if level == nil then level = 0 end
  if level > MAX_LEVEL then
      logError("max depth reached in p : "..tostring(MAX_LEVEL))
      return ""
  end
  local text = ""
  if (type(o) == "table") then
      text = "\n"
      for key,value in pairs(o) do
          for i=0, level do
              text = text .. " "
          end
          text = text .. ".".. key.."="..p(value, level+1) .. "\n"
      end
  elseif (type(o) == "function") then
      text = "[function]"
  elseif (type(o) == "boolean") then
      if o == true then
          text = "[true]"
      else
          text = "[false]"
      end
  else
      if o == nil then
          text = "[nil]"
      else
          text = tostring(o)
      end
  end
  return text
end

local function p(o, level)
  if o and type(o) == "table" and (o.x and o.z and o.y and #o == 3) then
      return string.format("{x=%s, z=%s, y=%s}", p(o.x), p(o.z), p(o.y))
  elseif o and type(o) == "table" and (o.x and o.y and #o == 2)  then
      return string.format("{x=%s, y=%s}", p(o.x), p(o.y))
  end
  return _p(o, level)
end

---checks if a string starts with a prefix
---@param aString any
---@param aPrefix any
---@param caseSensitive? boolean   ; if true, case sensitive search
---@return boolean
local function startsWith(aString, aPrefix, caseSensitive)
    local aString = aString
    if not aString then
        return false
    elseif not caseSensitive then
        aString = aString:upper()
    end
    local aPrefix = aPrefix
    if not aPrefix then
        return false
    elseif not caseSensitive then
        aPrefix = aPrefix:upper()
    end
    return string.sub(aString,1,string.len(aPrefix))==aPrefix
end

local function deepcopy(obj, seen)
    -- Handle non-tables and previously-seen tables.
    if type(obj) ~= 'table' then return obj end
    if seen and seen[obj] then return seen[obj] end

    -- New table; mark it as seen and copy recursively.
    local s = seen or {}
    local res = {}
    s[obj] = res
    for k, v in pairs(obj) do res[deepcopy(k, s)] = deepcopy(v, s) end
    return setmetatable(res, getmetatable(obj))
end

-------------------------------------------------------------------------------------------------------------------------------------------------------------
-- Core methods
-------------------------------------------------------------------------------------------------------------------------------------------------------------

---Loads the .lua settings file containing all the spawnable aircrafts
---@param settingsPath string the complete path to the settings file
---@param nameFilter string? a regex that will filter out the group names that we want to load; defaults to nil -> all groups are loaded
---@return table veafSpawnableAircrafts a table containing all the loaded spawnable aircraft groups, by lowercased name
local function loadSpawnableAircraftSettings(settingsPath, nameFilter)
  veafSpawnableAircraftsEditor.logDebug(string.format("Loading settings from [%s]",settingsPath))
  local file, err = loadfile(settingsPath)
  if file then
      file()
  else
    veafSpawnableAircraftsEditor.logInfo(string.format("File does not exist: [%s]",settingsPath))
  end

  local veafSpawnableAircrafts = {}

  local function addGroupIfFilterMatches(categoryName, coalitionName, countryName, groupData)
    local groupName = groupData.name or ""
    if (nameFilter == nil) or groupName:match(nameFilter) then
      groupData.veafSpawnableAircraftsEditorData = {}
      groupData.veafSpawnableAircraftsEditorData.categoryName = categoryName
      groupData.veafSpawnableAircraftsEditorData.coalitionName = coalitionName
      groupData.veafSpawnableAircraftsEditorData.countryName = countryName
      veafSpawnableAircrafts[groupData.name:lower()] = groupData
    end
  end

  if settings then
    -- old syntax
    --[[
      settings = {
    ["FR planes"] = {
        ["category"] = "plane",
        ["coalition"] = "blue",
        ["country"] = "france",
        ["groups"] = {
            [1] = {
    ]]
    for _, settingData in pairs(settings) do
      local groups_t = settingData.groups
      if groups_t then
        for _, groupData in pairs(groups_t) do
          addGroupIfFilterMatches(settingData.category, settingData.coalition, settingData.country, groupData)
        end
      end
    end
    -- new syntax
    --[[
      settings = {
        ["categories"] = {
            ["plane"] = {
                ["coalitions"] = {
                    ["blue"] = {
                        ["countries"] = {
                            ["france"] = {
                                ["groups"] = {
                                    ["veafSpawn-m2000-fox1"] = {         
    ]]
    local categories_t = settings.categories
    if categories_t then
      for categoryName, categoryData in pairs(categories_t) do
        local coalitions_t = categoryData.coalitions
        if coalitions_t then
          for coalitionName, coalitionData in pairs(coalitions_t) do
            local countries_t = coalitionData.countries
            if countries_t then
              for countryName, countryData in pairs(countries_t) do
                local groups_t = countryData.groups
                if groups_t then
                  for _, groupData in pairs(groups_t) do
                    addGroupIfFilterMatches(categoryName, coalitionName, countryName, groupData)
                  end
                end
              end
            end
          end
        end
      end
    end
    veafSpawnableAircraftsEditor.logDebug("Settings loaded")
  end

  return veafSpawnableAircrafts
end

local function processGroup(group)
  local rGroup = nil
  if group ~= nil then
    rGroup = deepcopy(group)
    if rGroup.name and not startsWith(rGroup.name, "veafSpawn-", false) then
      rGroup.name = "veafSpawn-" .. rGroup.name
    end
    rGroup.hidden = true
    rGroup.lateActivation = true
    rGroup.x = DEFAULT_POSITION_X
    rGroup.y = DEFAULT_POSITION_Y
    local route = rGroup.route
    if route then
      local points = route.points
      if points then
        for index, point in pairs(points) do
          if index ~= 1 then
            points[index] = nil -- remove all navigation points after the first one
          else
            points[index].x = DEFAULT_POSITION_X
            points[index].y = DEFAULT_POSITION_Y
          end
        end
      end
    end
    local units = rGroup.units
    if units then
      for _, unit in pairs(units) do
        if unit.name and not startsWith(unit.name, "veafSpawn-", false) then
          unit.name = "veafSpawn-" .. unit.name
        end
        unit.x = DEFAULT_POSITION_X
        unit.y = DEFAULT_POSITION_Y
      end
    end
  end
  return rGroup
end

---Reads a mission file and appends the resulting groups (after processing) into the spawnable aircrafts settings file
---@param missionPath string the mission file path (folder where the `mission` file resides)
---@param settingsPath string the spawnable aircrafts settings file path
---@param nameFilter string? a regex that will filter out the group names that we want to get from the mission; defaults to nil -> all groups are loaded
function veafSpawnableAircraftsEditor.retrieveFromMission(missionPath, settingsPath, nameFilter)

  -- load the mission file
  veafSpawnableAircraftsEditor.logDebug(string.format("Loading mission from [%s]",missionPath))
  local filePath = missionPath .. "\\mission"
  local file = assert(loadfile(filePath))
  if not file then
    veafMissionEditor.logError(string.format("Error while loading mission file [%s]",missionPath))
    return
  end
  file()
---@diagnostic disable-next-line: undefined-global -- "mission" is defined by the file() function
  local mission = mission
  veafSpawnableAircraftsEditor.logDebug("Mission loaded")

  local function processCategory(missionGroups, coalitionName, countryName, countryData, categoryName)
    local category_t = countryData[categoryName]
    if category_t then
      veafSpawnableAircraftsEditor.logDebug(string.format("  checking category [%s]", categoryName))
      local group_t = category_t.group
      for _, groupData in pairs(group_t) do
        local groupName = groupData.name
        veafSpawnableAircraftsEditor.logDebug(string.format("   groupName=%s",p(groupName)))
        if groupName then
          if not(nameFilter) or groupName:match(nameFilter) then
            veafSpawnableAircraftsEditor.logDebug("   Group name checked")
            local missionGroup = processGroup(groupData)
            missionGroup.veafSpawnableAircraftsEditorData = {}
            missionGroup.veafSpawnableAircraftsEditorData.coalitionName = coalitionName
            missionGroup.veafSpawnableAircraftsEditorData.countryName = countryName
            missionGroup.veafSpawnableAircraftsEditorData.categoryName = categoryName
            table.insert(missionGroups, missionGroup)
          end
        end
      end
    end
  end

  -- parse the mission and get the aircrafts groups
  local missionGroups = {}
  if mission then
    local coalition_t = mission.coalition
    if coalition_t then
      for coalitionName, coalitionData in pairs(coalition_t) do
        veafSpawnableAircraftsEditor.logDebug(string.format("  coalitionName=%s",p(coalitionName)))
        local country_t = coalitionData.country
        if country_t then
          for _, countryData in pairs(country_t) do
            local countryName = countryData.name
            if countryName then
              veafSpawnableAircraftsEditor.logDebug(string.format("  countryName=%s",p(countryName)))
              -- process categories
              for _, categoryName in pairs(CATEGORIES) do
                processCategory(missionGroups, coalitionName, countryName, countryData, categoryName)
              end
            end
          end
        end
      end
    end
  end

  -- load the spawnable aircrafts settings file
  local veafSpawnableAircrafts = loadSpawnableAircraftSettings(settingsPath) -- load all groups, do not apply nameFilter!

  -- inject the mission groups into the spawnable aircrafts groups
  if missionGroups then
    for _, groupData in pairs(missionGroups) do
      local groupName = groupData.name
      if veafSpawnableAircrafts[groupName:lower()] then
        veafSpawnableAircraftsEditor.logInfo(string.format("REPLACING group [%s]", groupName))
      else
        veafSpawnableAircraftsEditor.logInfo(string.format("ADDING group [%s]", groupName))
      end
      veafSpawnableAircrafts[groupName:lower()] = groupData -- replace or insert
    end
  end

  -- save the spawnable aircrafts settings file
  local rSettings = { categories = {}}
  local rCategories = rSettings.categories
  for _, groupData in pairs(veafSpawnableAircrafts) do
    local categoryName = groupData.veafSpawnableAircraftsEditorData.categoryName:lower()
    if not rCategories[categoryName] then
      rCategories[categoryName] = { coalitions = {}}
    end
    local rCoalitions = rCategories[categoryName].coalitions
    local coalitionName = groupData.veafSpawnableAircraftsEditorData.coalitionName:lower()
    if not rCoalitions[coalitionName] then
      rCoalitions[coalitionName] = { countries = {}}
    end
    local rCountries = rCoalitions[coalitionName].countries
    local countryName = groupData.veafSpawnableAircraftsEditorData.countryName
    if not rCountries[countryName] then
      rCountries[countryName] = { groups = {}}
    end
    local rGroups = rCountries[countryName].groups
    groupData.veafSpawnableAircraftsEditorData = nil
    local groupName = groupData.name
    rGroups[groupName] = groupData
  end
  local tableAsLua = veafMissionEditor.serialize("settings", rSettings, nil, false)
  veafMissionEditor.writeMissionFile(settingsPath, tableAsLua)
end

function veafSpawnableAircraftsEditor.injectInMission(filePath, settingsPath, nameFilter)
  -- load the spawnable aircrafts settings file
  local veafSpawnableAircrafts = loadSpawnableAircraftSettings(settingsPath, nameFilter) -- only load groups that match the nameFilter parameter, if any

  -- map the spawnable aircrafts groups by country and category
  local veafSpawnableAircraftsByCountryAndCategory = {}
  for groupName, groupData in pairs(veafSpawnableAircrafts) do
    local countryName = groupData.veafSpawnableAircraftsEditorData.countryName
    if countryName then
      if veafSpawnableAircraftsByCountryAndCategory[countryName] == nil then
        veafSpawnableAircraftsByCountryAndCategory[countryName] = {}
      end
      local countryCategoriesTable = veafSpawnableAircraftsByCountryAndCategory[countryName]
      local categoryName = groupData.veafSpawnableAircraftsEditorData.categoryName
      if countryCategoriesTable[categoryName] == nil then
        countryCategoriesTable[categoryName] = {}
      end
      countryCategoriesTable[categoryName][groupName] = groupData
    end
  end

  local maxGroupId = 1
  local maxUnitId = 1

  local function computeMaxIds(o)
    if (type(o) == "table") then
      for key,value in pairs(o) do
        veafSpawnableAircraftsEditor.logTrace(string.format("parseTable %s", p(key)))
        if tostring(key):lower() == "groupid" then
          maxGroupId = maxGroupId + 1
          o[key] = maxGroupId
          veafSpawnableAircraftsEditor.logDebug(string.format("maxGroupId=[%s]", p(maxGroupId)))
        elseif tostring(key):lower() == "unitid" then
          maxUnitId = maxUnitId + 1
          o[key] = maxUnitId
          veafSpawnableAircraftsEditor.logDebug(string.format("maxUnitId=[%s]", p(maxUnitId)))
        end
        computeMaxIds(value)
      end
    end
  end

  ---This function is called by the mission editor to process the lua mission table before it's written back to disk
  ---@param missionTable table a LUA mission table as read from the `mission` file
  ---@return table missionTable the same LUA mission table, processed and ready to be written back to the `mission` file
  local function processMissionTable(missionTable)

    -- find max maxGroupId and unitId
    computeMaxIds(missionTable) -- this will populate the local maxUnitId and maxGroupId variables
    veafSpawnableAircraftsEditor.logDebug(string.format("maxGroupId=[%s]", p(maxGroupId)))
    veafSpawnableAircraftsEditor.logDebug(string.format("maxUnitId=[%s]", p(maxUnitId)))
  
    -- find the country tables in the mission table
    local missionCountryTablesByName = {}
    local missionCoalitionTablesByName = {}
    local missionCoalitionsNamesByCountryName = {} -- TODO pas sûr que ça serve
    -- browse coalitions
    for _, missionCoalition_t in pairs(missionTable["coalition"]) do
      local missionCoalitionName = missionCoalition_t["name"]
      veafSpawnableAircraftsEditor.logTrace(string.format("found coalition [%s]",missionCoalitionName))
      if missionCoalitionTablesByName[missionCoalitionName] == nil then
        missionCoalitionTablesByName[missionCoalitionName] = missionCoalition_t
      end
      -- browse countries
      for _, missionCountry_t in pairs(missionCoalition_t["country"]) do
        local missionCountryName = missionCountry_t["name"]
        if missionCountryName then
          veafSpawnableAircraftsEditor.logTrace(string.format("found country [%s]",missionCountryName))
          missionCoalitionsNamesByCountryName[missionCountryName] = missionCoalitionName
          missionCountryTablesByName[missionCountryName:lower()] = missionCountry_t
        end
      end
    end

    local currentGroupId = maxGroupId
    local currentUnitId = maxUnitId

    -- process all the spawnable aircrafts groups by country
    for spawnCountryName, spawnCategoriesTable in pairs(veafSpawnableAircraftsByCountryAndCategory) do
      local missionCountry_t = missionCountryTablesByName[spawnCountryName]
      local missionCoalitionName = missionCoalitionsNamesByCountryName[spawnCountryName]
      -- create the country in the mission if missing
      if not missionCountry_t then
        -- TBD later, I need the link between country name and country ID
        veafSpawnableAircraftsEditor.logError(string.format("Missing country %s in the mission; please add this country to the mission before trying again", spawnCountryName))
      else
        veafSpawnableAircraftsEditor.logTrace(string.format("Processing country [%s]", spawnCountryName))
        -- process categories
        for spawnCategoryName, spawnGroupsTable in pairs(spawnCategoriesTable) do
          local missionCategory_t = missionCountry_t[spawnCategoryName]
          -- create the category if it doesn't exist
          if not missionCategory_t then
            veafSpawnableAircraftsEditor.logInfo(string.format("Missing category %s for country %s in the mission; adding it automatically", spawnCategoryName, spawnCountryName))
            missionCountry_t[spawnCategoryName] = { group = {}}
          end
          for spawnGroupName, spawnGroupData in pairs(spawnGroupsTable) do
            -- find the group in the category groups table
            local missionGroups_t = missionCountry_t[spawnCategoryName].group
            local foundIndex = nil
            for index, group_t in pairs(missionGroups_t) do
              local missionGroupName = group_t.name
              if missionGroupName == spawnGroupName then
                foundIndex = index
                break
              end
            end
            local newGroupData = deepcopy(spawnGroupData)
            -- change the group and units ids
            currentGroupId = currentGroupId + 1
            newGroupData.groupId = currentGroupId
            for _, newUnitData in pairs(newGroupData.units) do
              currentUnitId = currentUnitId + 1
              newUnitData.unitId = currentUnitId
            end
            -- reset work data
            newGroupData.veafSpawnableAircraftsEditorData = nil
            if foundIndex then
              -- found it, let's replace the group data
              missionGroups_t[foundIndex] = newGroupData
            else
              -- inject the group as new
              table.insert(missionGroups_t, newGroupData)
            end
          end
        end
      end
    end

    return missionTable
  end

  -- edit the "mission" file
  veafSpawnableAircraftsEditor.logDebug(string.format("Processing mission at [%s]",filePath))
  local _filePath = filePath .. "\\mission"
  veafMissionEditor.editMission(_filePath, _filePath, "mission", processMissionTable)
  veafSpawnableAircraftsEditor.logDebug("Mission edited")
end

veafSpawnableAircraftsEditor.logDebug(string.format("#arg=%d",#arg))
for i=0, #arg do
    veafSpawnableAircraftsEditor.logDebug(string.format("arg[%d]=%s",i,arg[i]))
end
if #arg < 2 then
    veafSpawnableAircraftsEditor.logError("USAGE : veafSpawnableAircraftsEditor.lua <mission folder path> <settings file> [-debug|-trace] [-import] [-namefilter <filter>]")
    return
end

veafSpawnableAircraftsEditor.debug = false
veafSpawnableAircraftsEditor.trace = false
local import = false
local nameFilter = nil
local filePath = arg[1]
local settingsPath = arg[2]
for i = 3, #arg, 1 do
  -- processing an argument
  if arg[i]:lower() == "-debug" then
    veafSpawnableAircraftsEditor.debug = true
  elseif arg[i]:lower() == "-trace" then
    veafSpawnableAircraftsEditor.trace = true
  elseif arg[i]:lower() == "-import" then
    import = true
  elseif arg[i]:lower() == "-namefilter" then
    nameFilter = arg[i+1]
  end
end
if veafSpawnableAircraftsEditor.debug or veafSpawnableAircraftsEditor.trace then
  veafSpawnableAircraftsEditor.Debug = true
  veafMissionEditor.Debug = true
  if veafSpawnableAircraftsEditor.trace then
    veafSpawnableAircraftsEditor.Trace = true
    veafMissionEditor.Trace = true
  end
else
  veafSpawnableAircraftsEditor.Debug = false
  veafMissionEditor.Debug = false
  veafSpawnableAircraftsEditor.Trace = false
  veafMissionEditor.Trace = false
end

if import then
  veafSpawnableAircraftsEditor.logInfo("Importing data from the mission and/or reprocessing the settings file)")
  veafSpawnableAircraftsEditor.retrieveFromMission(filePath, settingsPath, nameFilter)
else
  veafSpawnableAircraftsEditor.logInfo("Injecting data into the mission")
  veafSpawnableAircraftsEditor.injectInMission(filePath, settingsPath, nameFilter)
end